#
# ENTER THE STRING
#
# My metatheme for this competition is "quick-n-dirty".
# (I usually tend towards astraction and "elegant" design -
# not this week.)  I need this metatheme - the official
# theme isn't working for me.
#
# pyweek 7 entry
# author: Drew Smathers
#

import math
import random

import euclid

from pyglet import options
options['vsync'] = 0
options['gl_debug'] = 0

from pyglet.gl import *
import pyglet
from pyglet.window import key, mouse

import miru
from miru.ext import geom, debug
from miru import effects
from miru.ui import Window

DEBUG = False

context = miru.context

context.window = window = Window(400, 600,
        caption='Enter the String')
if DEBUG:
    miru.utils.add_fps_display(context)
context.camera.lights = miru.camera.LightGroup([
             miru.camera.PositionalLight(pos=(2,0,5), kq=0.4),
             miru.camera.PositionalLight(pos=(-2,1,0), kq=0.5),
             miru.camera.PositionalLight(pos=(1.5,1.5,-3), kq=0.6),
             miru.camera.PositionalLight(pos=(0,2,-5), kq=0.7),
             miru.camera.PositionalLight(pos=(1.5,1.5,2), kq=0.5),
    ])
context.camera.pos = (0,7.5,4)
context.camera.rotation_mode = context.camera.ROTATE_MODE
context.camera.angle.x = 55
#context.camera.depth_sort = True

# Set up global line attributes
glLineWidth(2)
glEnable(GL_LINE_STIPPLE)
glLineStipple(1, 0x55555)

class MouseControl(miru.input.MouseControl):

    def on_select(self, target, x, y, button, modifiers):
        if hasattr(target, 'bead'):
            if target.bead.animating:
                return
            if keyboard[key.SPACE] or (modifiers & key.MOD_CTRL):
                beadSwapper.push(target.bead, repush=True)
            elif button == mouse.LEFT:
                if beadSwapper.stacked:
                    beadSwapper.push(target.bead, repush=False)
                else:
                    target.bead.rotate()
            elif button == mouse.RIGHT:
                beadSwapper.push(target.bead, repush=True)

    def on_mouse_scroll(self, x, y, dx, dy):
        pass
    
    def on_mouse_release(self, x, y, button, modifiers):
        pass

    def on_mouse_drag(self, x, y, dx, dy, button, modifiers):
        pass

# Track keyboard state
keyboard = key.KeyStateHandler()
window.push_handlers(keyboard)

context.control = MouseControl()

class Block(object):
    indexed = False
    vertex_pitch = 3
    drawing_mode = gl.GL_QUADS

    def __init__(self, width, height, length):
        w, h, l = width / 2., height / 2., length / 2.
        fbl = (-w, -h, l)
        ftl = (-w, h, l)
        btl = (-w, h, -l)
        bbl = (-w, -h, -l)
        
        bbr = (w, -h, -l)
        btr = (w, h, -l)
        ftr = (w, h, l)
        fbr = (w, -h, l)
        self.vertex_data = geom._flatten([
            fbl, ftl, btl, bbl, # left side
            bbl, btl, btr, bbr, # back 
            bbr, btr, ftr, fbr, # right side
            btl, ftl, ftr, btr, # top
            fbl, bbl, bbr, fbr, # bottom
            ftl, fbl, fbr, ftr # front
        ])
        self.normal_data = [
            -1,0,0,   -1,0,0,  -1,0,0,  -1,0,0,
            0,0,-1,   0,0,-1,  0,0,-1,  0,0,-1,
            1,0,0, 1,0,0, 1,0,0, 1,0,0,
            0,1,0, 0,1,0,  0,1,0,  0,1,0,
            0,-1,0,  0,-1,0,  0,-1,0, 0,-1,0,
            0,0,1, 0,0,1, 0,0,1, 0,0,1]
        self.texcoord_data = [
            0.0, 0.0,  1.0, 0.0,  1.0, 1.0,  0.0, 1.0
            ] * 6

group_white = miru.graphics.ColorGroup((1,1,1,1))
group_red = miru.graphics.ColorGroup((1,0,0,1))
group_green = miru.graphics.ColorGroup((0,1,0,1))
group_blue = miru.graphics.ColorGroup((0,0,1,1))
group_yellow = miru.graphics.ColorGroup((1,1,0,1))
group_brown = miru.graphics.ColorGroup((0.8,0.8,0.3,1))
group_charcoal = miru.graphics.ColorGroup((0.15,0.05,0.15,1))
group_ablack = miru.graphics.ColorGroup((0.0,0.0,0.0,0.65))

class Bead:

    _batch = None
    _cache = {}
    _lines_cache = {}
    _stars_cache = None
    _stars2_cache = None

    RED = 0
    GREEN = 1
    BLUE = 2
    YELLOW = 3
    WHITE = 4

    NORTH = 1
    EAST = 2
    SOUTH = 4
    WEST = 8

    def __init__(self, faces=0xf, bead_type=None, row=None):

        self.bead_type = bead_type
        self.faces = faces
        self.row = row
        self.north_row = None

        # evaluated and linked count used in final evaluation
        self.evaluated = False
        self.evaluator_root = None
        self._linked_count = [0]

        self.obj = None
        self.east_node = None
        self.south_node = None
        self.west_node = None
        self.north_node = None
        
        self.lines_batch = None
        self.stars_batch = None
        
        self.animating = False
        self.offset = euclid.Vector3(0,0,0)

        if bead_type is None:
            self.bead_type = Bead.RED
        self.rebatch()

    def rebatch(self):
        key = (self.faces, self.bead_type)
        if key in Bead._cache:
            self.batch = Bead._cache[key]
            if not self.obj:
                self.obj = miru.core.Object(self)
                self.obj.bead = self
            #else:
            #    self.obj.drawable = self.batch
        else:
            self._make_batch(key)

    def _make_batch(self, key):
        w = 0.25
        batch = pyglet.graphics.Batch()
        heart = geom.Sphere(radius=w*0.45, slices=8, stacks=8)
        y = w * 0.5
        h = 0
        if self.faces & Bead.NORTH:
            block = Block(w * 2, y, 0.05)
            geom.transform(block, (0, -y, -w))
            geom.get_vlist(block, batch, group_white)
        if self.faces & Bead.SOUTH:
            block = Block(w * 2, y, 0.05)
            geom.transform(block, (0, -y, w))
            geom.get_vlist(block, batch, group_white)
        if self.faces & Bead.EAST:
            block = Block(0.05, y, w * 2)
            geom.transform(block, (-w, -y, 0))
            geom.get_vlist(block, batch, group_white)
        if self.faces & Bead.WEST:
            block = Block(0.05, y, w * 2)
            geom.transform(block, (w, -y, 0))
            geom.get_vlist(block, batch, group_white)
        group = {
            Bead.RED: group_red,
            Bead.GREEN: group_green,
            Bead.BLUE: group_blue,
            Bead.YELLOW: group_yellow,
            Bead.WHITE: group_white,
        }[self.bead_type]
        # add the base
        batch.add(4, GL_QUADS, group_charcoal,
                ('n3f', (0,1,0,0,1,0,0,1,0,0,1,0)),
                ('v3f', (-w, -y, -w,
                          w, -y, -w,
                          w, -y, w,
                         -w, -y, w)))
        # and the heart ...
        geom.get_vlist(heart, batch, group)
        self.batch = batch
        if not self.obj:
            self.obj = miru.core.Object(self)
            self.obj.bead = self
        #else:
        #    self.obj.drawable = batch
        Bead._cache[key] = batch


    def draw(self):
        if self.offset:
            #self.obj.pos += self.offset
            #o = self.offset * 0.5
            glTranslatef(*self.offset)
        self.batch.draw()
        if self.lines_batch:
            self.lines_batch.draw()
        if self.stars_batch:
            self.stars_batch.draw()
        #if self.offset:
        #    self.obj.pos -= self.offset

    def rotate(self, direction='ccw'):
        if direction == 'ccw':
            self.faces <<= 1
            self.faces += self.faces >> 4
            self.faces &= 0xf
        else:
            tmp = self.faces
            self.faces >>= 1
            self.faces |= tmp << 3
        self._faces = self.faces
        self.faces = Bead.EAST | Bead.SOUTH | Bead.WEST | Bead.NORTH
        self._update_neighbors()
        self.animating = True
        pyglet.clock.schedule_interval(self._update_rotation, 1/60.)


    def _update_neighbors(self):
        for bead in self.row:
            bead.rebatch_lines()
        if self.north_row:
            for bead in self.north_row:
                bead.rebatch_lines()

    def _update_rotation(self, dt):
        if not self.animating:
            return
        dy = 500 * dt
        self.obj.angle.y += dy
        if self.obj.angle.y > 90:
            self.faces = self._faces
            pyglet.clock.unschedule(self._update_rotation)
            self.animating = False
            self.obj.angle.y = 0
            self.rebatch()
            self._update_neighbors()


    def connected_east(self):
        return ( self.east_node and
                 (( self.bead_type == Bead.WHITE or 
                    self.east_node.bead_type == Bead.WHITE ) or
                   self.bead_type == self.east_node.bead_type) and
                 not (Bead.EAST & self.faces) and
                 not (Bead.WEST & self.east_node.faces) )

    def connected_south(self):
        return ( self.south_node and
                 (( self.bead_type == Bead.WHITE or 
                    self.south_node.bead_type == Bead.WHITE ) or
                   self.bead_type == self.south_node.bead_type) and
                 not (Bead.SOUTH & self.faces) and
                 not (Bead.NORTH & self.south_node.faces) )

    def rebatch_lines(self):

        if self.connected_east() and self.connected_south():
            key = (1, 1, self.bead_type)
        elif self.connected_east():
            key = (1, 0, self.bead_type)
        elif self.connected_south():
            key = (0, 1, self.bead_type)
        else:
            self.lines_batch = None
            return

        self.lines_batch = Bead._lines_cache.get(key)
        if not self.lines_batch:
            self.lines_batch = batch = pyglet.graphics.Batch()
            group = {
                Bead.RED: group_red,
                Bead.GREEN: group_green,
                Bead.BLUE: group_blue,
                Bead.YELLOW: group_yellow,
                Bead.WHITE: group_white,
            }[self.bead_type]
            y = 0.05
            x = 0.05
            if key[0]:
                batch.add(4, GL_LINES, group,
                        (('v3f'),
                            (0, y, -x, -0.75, y, -x,
                             0, y, x, -0.75, y, x)
                            ))
            if key[1]:
                batch.add(4, GL_LINES, group,
                        (('v3f'),
                            (-x, y, 0, -x, y, 0.75,
                             x, y, 0, x, y, 0.75)))
            Bead._lines_cache[key] = batch
       
    def add_stars(self, surround=True):
        self.stars_batch = Bead._stars_cache
        if not self.stars_batch:
            self.stars_batch = batch = pyglet.graphics.Batch()
            data = halo_data(count=11)
            vlist = batch.add(11, GL_POINTS, group_stars_cyan, ('v3f', data))
            Bead._stars_cache = batch
            animation = HaloAnimation(vlist, count=11)
            pyglet.clock.schedule_interval(animation.update, 1/60.)
        if not surround:
            return
        if hasattr(self.east_node, 'add_stars2'):
            self.east_node.add_stars2()
        if hasattr(self.south_node, 'add_stars2'):
            self.south_node.add_stars2()
        if hasattr(self.west_node, 'add_stars2'):
            self.west_node.add_stars2()
        if hasattr(self.north_node, 'add_stars2'):
            self.north_node.add_stars2()

    def remove_stars(self):
        self.stars_batch = None
        if self.east_node:
            self.east_node.stars_batch = None
        if self.south_node:
            self.south_node.stars_batch = None
        if self.west_node:
            self.west_node.stars_batch = None
        if self.north_node:
            self.north_node.stars_batch = None


    def do_swap(self, other):
        # Start the swap animation
        self.animating = True
        other.animating = True
        self._swap_neighbor = other
        self._swap_theta = 0
        self.lines_batch = None
        other.lines_batch = None
        self.add_stars(False)
        other.add_stars(False)

        if self.south_node == other:
            # 0.5 * 0.25
            # self._swap_fulcrum = self.obj.pos.z + 0.5 * (other.obj.pos.z - self.obj.pos.)z
            print 'starting verticle swap'
            pyglet.clock.schedule_interval(self._update_swap_vert, 1/60.)
        else:
            #self._swap_fulcrum = self.obj.pos.x + 
            pyglet.clock.schedule_interval(self._update_swap_horz, 1/60.)
        

    def on_swap_complete(self, f):
        other = self._swap_neighbor

        self.offset = euclid.Vector3(0,0,0)
        other.offset = euclid.Vector3(0,0,0)

        self.animating = False
        other.animating = False
        pyglet.clock.unschedule(f)
        self.stars_batch = None
        other.stars_batch = None

        tmp_faces = self.faces
        tmp_bead_type = self.bead_type
        self.faces = other.faces
        self.bead_type = other.bead_type
        other.faces = tmp_faces
        other.bead_type = tmp_bead_type

        self.rebatch()
        self._update_neighbors()
        other.rebatch()
        other._update_neighbors()

    def _update_swap_vert(self, dt):
        other = self._swap_neighbor
        radius = abs(other.obj.pos.z - self.obj.pos.z) * 0.5

        dtheta = 10 * dt
        self._swap_theta += dtheta

        if self._swap_theta > math.pi:
            return self.on_swap_complete(self._update_swap_vert)


        theta = self._swap_theta
        self.offset.y = radius * math.sin(math.pi + theta)
        self.offset.z = radius * math.cos(math.pi + theta) + radius
        other.offset.y = radius * math.sin(theta)
        other.offset.z = radius * math.cos(theta) - radius
            

    def _update_swap_horz(self, dt):
        other = self._swap_neighbor
        radius = abs(other.obj.pos.x - self.obj.pos.x) * 0.5

        dtheta = 10 * dt
        self._swap_theta += dtheta

        if self._swap_theta > math.pi:
            return self.on_swap_complete(self._update_swap_horz)


        theta = self._swap_theta
        self.offset.y = radius * math.sin(theta)
        self.offset.x = radius * math.cos(theta) - radius
        other.offset.y = radius * math.sin(math.pi + theta)
        other.offset.x = radius * math.cos(math.pi + theta) + radius
    
    def add_stars2(self):
        self.stars_batch = Bead._stars2_cache
        if not self.stars_batch:
            self.stars_batch = batch = pyglet.graphics.Batch()
            data = halo_data(radius=0.17, count=6)
            vlist = batch.add(6, GL_POINTS, group_stars_yellow, ('v3f', data))
            Bead._stars2_cache = batch
            animation = HaloAnimation(vlist, radius=0.17, count=6)
            pyglet.clock.schedule_interval(animation.update, 1/60.)
    
    @property
    def linked_count(self):
        return self._linked_count[0]

    def evaluate(self, start=True, ct=(), root=None):

        if self.evaluated:
            return
        self.evaluated = True

        if root == None:
            self.evaluator_root = root = self
        if not ct:
            ct = [1]
        else:
            ct[0] += 1
        self._linked_count = ct
        #ct[0] += 1
        # Add two for the current node if not white
        if self.bead_type != Bead.WHITE and\
           (self.connected_south() or self.connected_east() or\
            (self.north_node and self.north_node.connected_south()) or\
            root != self):
            context.statisticsBoard.add_bead(self.bead_type)
        if self.connected_south():
            # and 1 for the link
            context.statisticsBoard.add_link()
            # visit the south node
            self.south_node.evaluate(start, ct, root)
        if self.connected_east():
            context.statisticsBoard.add_link()
            # visit the east node
            self.east_node.evaluate(start, ct, root)
        #print 'linked count', self.linked_count

    def finalize_evaluation(self):
        print 'finalize linked count', self.linked_count
        bonuses5 = self.linked_count // 5
        bonuses15 = self.linked_count // 15
        for i in range(bonuses5):
            context.statisticsBoard.add_bonus5()
        for i in range(bonuses15):
            context.statisticsBoard.add_bonus15()
        if bonuses5:
            print '5+ bonus!', bonuses5
        #context.statisticsBoard.fifteen_bonuses += bonuses15
        if bonuses15:
            print '15+ bonus!', bonuses15
        #TODO - generate woo-hoo events for bonuses
        print '(%r) current total : %d' % (self, context.statisticsBoard.total)


def halo_data(theta_start=0, radius=0.23, y=0.05, count=11):
    theta = theta_start
    incr = math.pi * 2. / count
    data = []
    for i in range(count):
        x = radius * math.cos(theta)
        z = radius * math.sin(theta)
        data.append([x,y,z])
        theta += incr
    data.sort(lambda a,b: cmp(a[2], b[2]))
    data = miru.utils.flatten(data)
    return data

class HaloAnimation:

    def __init__(self, halo_vlist, radius=0.23, count=11):
        self.halo_vlist = halo_vlist
        self.angle_velocity = 2
        self.theta = 0
        self.count = count
        self.radius = radius

    def update(self, dt):
        dtheta = dt * self.angle_velocity
        self.theta += dtheta
        data = halo_data(self.theta, radius=self.radius, count=self.count)
        self.halo_vlist.vertices[:] = data


class PseudoBead:

    def __init__(self, linked_count, bead_type, faces, root, visit=()):
        self.linked_count = linked_count
        self.bead_type = bead_type
        self.faces = faces
        self.evaluator_root = root
        self.evaluated = True
        self.visited = False
        self.visit = visit

    def evaluate(self, start, ct, root):
        if self.visited:
            return
        ct[0] += self.linked_count
        self.evaluator_root = root
        self.visited = True
        for node in self.visit:
            node.evaluate(ct=ct, root=root)

    def finalize_evaluation(self):
        return

class BeadStream:

    face_configuration = [0,0] + (range(0xf) * 7)
    bead_configuration = ( [Bead.RED] * 3 +
                           [Bead.GREEN] * 3 +
                           [Bead.BLUE] * 3 +
                           [Bead.YELLOW] * 3 +
                           [Bead.WHITE] )

    def __init__(self, columns=3, velocity=0.23, stream_count=50):
        self.columns = columns
        self.velocity = 4.25
        self._velocity = velocity
        pyglet.clock.schedule_once(self.reset_veloctity, 1)
        self.rows = []
        self.count = 0
        self.stream_count = stream_count
        self.end_buffer = []
        self.advance()

    def reset_veloctity(self, dt):
        self.velocity = self._velocity

    def advance(self):

        if self.completed:
            return

        z = -5
        step = 0.75
        cur = -(step * self.columns * 0.5) + step * 0.5
        row = []
        for i in range(self.columns):
            face = random.choice(self.face_configuration)
            bead_type = random.choice(self.bead_configuration)
            #if i == self.columns - 1:
            #    face = Bead.WEST
            #    bead_type = Bead.RED
            #face = random.choice([Bead.NORTH, Bead.WEST])
            #bead_type = Bead.RED
            bead = Bead(face, bead_type, row)
            row.append(bead)
            bead.obj.pos = (cur, 0.5, -8)
            context.add_object(bead.obj)
            # XXX stupid hack to push object to front in rendering
            # order - add_object() function should probably allow for
            # specifying an index.
            added = context.camera.objects.pop()
            context.camera.objects.insert(1, added)
            # end hack
            cur += step
        if self.rows:
            last_row = self.rows[-1]
            for bead in last_row:
                bead.north_row = row
        else:
            last_row = [None] * self.columns
        self.rows.append(row)
        east_node = None
        for (north_node, south_node) in zip(row, last_row):
            if east_node:
                north_node.east_node = east_node
                east_node.west_node = north_node
            east_node = north_node
            north_node.south_node = south_node
            if south_node:
                south_node.north_node = north_node
        for bead in row:
            bead.rebatch_lines()
        self.count += 1


    @property
    def completed(self):
        return self.stream_count and self.count > self.stream_count

    def update(self, dt):
        dz = self.velocity * dt
        for row in self.rows:
            for bead in row:
                bead.obj.pos.z += dz
        if self.rows:
            tail = self.rows[-1][0]
            if tail.obj.pos.z > -7.2:
                self.advance()
                if len(self.rows) > 14:
                    front = self.rows.pop(0)
                    for bead in front:
                        context.remove_object(bead.obj)
                    self.end_buffer.append(front)
                    if self.end_buffer:
                        self.evaluate_front()
        if self.completed:
            if not self.rows and not self.end_buffer:
                return
            if not self.rows:
                while self.end_buffer:
                    self.evaluate_front()
                # TODO - generate level end event
                return
            front = self.rows[0]
            if front[0].obj.pos.z > 3.5:
                self.rows.pop(0)
                for bead in front:
                    context.remove_object(bead.obj)
                self.end_buffer.append(front)
                self.evaluate_front()
        #else:
        #    print '-' * 80
        #    print 'rows : ', len(self.rows)
        #    print 'completed : ', self.completed
    def evaluate_front(self):
        front = self.end_buffer.pop(0)
        # traverse right to left and evaluate beads
        roots = set()
        front.reverse()
        #print '-' * 80
        for bead in front:
            bead.evaluate()
            roots.add(bead.evaluator_root)
        collapsed = set()
        # Scan through middle row and possibly collapse any connected roots
        if self.rows:#self.end_buffer:
            head = self.rows[0]
            for bead in head:#self.end_buffer[0]:
                south_node = bead.south_node
                if bead.connected_south():
                    if south_node in roots:
                        bead.south_node = PseudoBead(south_node.linked_count,
                                south_node.bead_type, south_node.faces, bead)
                    else:
                        bead.south_node = PseudoBead(0,
                                south_node.bead_type, south_node.faces, bead,
                                visit=(bead,))
                    collapsed.add(south_node.evaluator_root)
        # Finalize roots if not collapsed
        for bead in roots:
            if not bead: # ???
                continue
            if bead not in collapsed:
                bead.finalize_evaluation()
                

class BeadSwapper:

    def __init__(self):
        self._stack = []

    @property
    def stacked(self):
        return bool(self._stack)

    def push(self, bead, repush=False):
        self._stack.append(bead)
        bead.add_stars()
        if len(self._stack) == 2:
            bead1 = self._stack.pop()
            bead1.remove_stars()
            bead2 = self._stack.pop()
            bead2.remove_stars()
            if bead2 == bead1.east_node or\
               bead2 == bead1.west_node or\
               bead2 == bead1.south_node or\
               bead2 == bead1.north_node:
                if bead1.east_node == bead2 or bead2.north_node == bead1:
                    bead1.do_swap(bead2)
                else:
                    bead2.do_swap(bead1)
                self.reset()
            else:
                # TODO - generate a doh! event
                self.reset()
                # Push new one back on
                if repush:
                    self.push(bead)

    def reset(self):
        for bead in self._stack:
            bead.remove_stars()
        self._stack = []

beadSwapper = BeadSwapper()


class StatisticsBoard:

    def __init__(self, pointsBlingger):
        # (x1)
        self.links = 0
        # (x2)
        self.nodes = 0
        # (x10)
        self.five_bonuses = 0
        # (x50)
        self.fifteen_bonuses = 0

        self.pointsBlingger = pointsBlingger

        self.level_totals = []

    @property
    def total(self):
        return self.links + self.nodes * 2 + self.five_bonuses * 10 +\
               self.fifteen_bonuses * 50

    def advance_level(self):
        self.level_totals.append(self.total)
        self.links = 0
        self.nodes = 0
        self.five_bonuses = 0
        self.fifteen_bonuses = 0

    def add_link(self):
        self.links += 1
        self.pointsBlingger.bling(1)

    def add_bead(self, bead_type):
        self.nodes += 1
        self.pointsBlingger.bling(2, bead_type)

    def add_bonus5(self):
        self.five_bonuses += 1
        self.pointsBlingger.bling(10)
    
    def add_bonus15(self):
        self.fifteen_bonuses += 1
        self.pointsBlingger.bling(50)

class PointsBlingger:

    def __init__(self):
        self._scheduled = []
        self.decay_rate = 0.8
        self.rise_velocity = 70.0
        self._queue = []

    def update(self, dt):
        dy = dt * self.rise_velocity
        dalpha = dt * self.decay_rate
        remove = []
        for label in list(self._scheduled):
            label.y += dy
            color = list(label.color)
            color[3] -= dalpha
            label.color = color
            if color[3] <= 0:
                self._scheduled.remove(label)
                context.remove_object(label, context.osd)
        

    def update_queue(self, dt):
        if not self._queue:
            return
        points, bead_type = self._queue.pop()
        size = { 1 : 20,
                 2 : 30,
                 10 : 50,
                 50 : 70,
                }[points]
        color = {
            Bead.RED: (1,0,0,1),
            Bead.GREEN: (0,1,0,1),
            Bead.BLUE: (0,0,1,1),
            Bead.YELLOW: (1,1,0,1),
            Bead.WHITE: (1,1,1,1),
        }[bead_type]

        # special cases for the bonuses
        if points == 10:
            color = (1,0,1,1)
        elif points == 50:
            color = (0,1,1,1)

        font = miru.utils.default_font(size=size)
        if points in (10, 50):
            x = 30 + random.random() * (context.window.width - 250)
        else:
            x = 30 + random.random() * (context.window.width - 90)
        y = 10
        label = pyglet.font.Text(font, "+%d" % points, x=x, y=y,
                color=color)
        context.add_object(label, context.osd)
        self._scheduled.append(label)
        #if self._queue:
        #    pyglet.clock.schedule_once(self.update_queue, 0.25)


    def bling(self, points, bead_type=Bead.WHITE):
        self._queue.append((points, bead_type))
        #if len(self._queue) == 1:
        #    pyglet.clock.schedule_once(self.update_queue, 0.25)

pointsBlingger = PointsBlingger()
context.statisticsBoard = StatisticsBoard(pointsBlingger)
pyglet.clock.schedule_interval(pointsBlingger.update, 1/60.)
pyglet.clock.schedule_interval(pointsBlingger.update_queue, 0.10)

class NoStippleGroup(pyglet.graphics.Group):

    def __init__(self, parent):
        super(NoStippleGroup, self).__init__(parent)

    def set_state(self):
        glPushAttrib( GL_ENABLE_BIT )
        glDisable( GL_LINE_STIPPLE )

    def unset_state(self):
        glPopAttrib()


class Hud:
    def __init__(self):

        group_nostipple = NoStippleGroup(group_brown)

        self.dashboard = pyglet.graphics.Batch()

        w, h = context.window.width, context.window.height
        self.dashboard.add(4, GL_QUADS, group_ablack,
            ('v2f', (0, h, 0, h - 40, w, h - 40, w, h)))
        self.dashboard.add(2, GL_LINES, group_nostipple,
            ('v2f', (0, h - 40, w, h - 40)))

        self._points_memo = context.statisticsBoard.total

        font = miru.utils.default_font(size=20)
        self.label = pyglet.font.Text(font, "Total 0", x=5, y=h-27)

    def draw(self):
        self.dashboard.draw()
        total = context.statisticsBoard.total
        if self._points_memo != total:
            self.label.text = 'Total %d' % total
            self._points_memo = total
        self.label.draw()

hud = Hud()
context.add_object(hud, context.osd)


# Make some fuzzy image data for particle effects
class FuzzyImagePattern(pyglet.image.ImagePattern):
    
    def __init__(self, color=(255,255,255)):
        self.color = '%c%c%c' % color

    def create_image(self, width, height):
        # ideally width should equal height
        radius = height // 2
        cx = width // 2
        cy = height // 2
        data = []
        for row in range(width):
            for col in range(height):
                dx = cx - row
                dy = cy - col
                d = math.sqrt(dx*dx + dy*dy)
                if d > radius:
                    alpha = 0
                else:
                    v = 1.0 - (d / radius)
                    v = v**3
                    alpha = int(min(255 * v, 255))
                color = self.color + '%c' % alpha
                data.append(color)
        data = ''.join(data)
        return pyglet.image.ImageData(width, height, 'RGBA', data)



# add reflective ground
#ground = make_ground()
#context.add_object(ground)

def make_ground():
    batch = pyglet.graphics.Batch()
    coords, tex = debug.checker_quad(32, scale=0.0078125,
            color1=(0.01,0.01,0.01,0.9), color2=(0,0,0,0.5))
    batch = pyglet.graphics.Batch()
    texgroup = miru.graphics.TextureBindGroup(tex)
    batch.add(4, GL_QUADS, texgroup,
            ('v3f', [-100, -2, -100, 100, -2, -100, 100, -2, 100, -100, -2, 100]),
            ('t4f', coords))
    ground = miru.core.Object(batch)
    return ground


#class Invisible:
#    def draw(self):
#        pass


#refl = effects.Reflection(camera=context.camera, ground=make_ground())
#context.camera.effects.append(refl)

# ---
context.add_object(make_ground())

beadStream = BeadStream(4)
pyglet.clock.schedule_interval(beadStream.update, 1/60.)
# TODO - abstract this into levels

# Experiment with point sprites
stars_white = pyglet.image.create(32,32, FuzzyImagePattern())
stars_yellow = pyglet.image.create(32, 32,
        FuzzyImagePattern(color=(255,255,0)))
stars_cyan = pyglet.image.create(32,32,
        FuzzyImagePattern(color=(0,255,255)))

group_stars_yellow = miru.graphics.PointSpriteGroup(
        stars_white.get_texture(), 32, attenuate=True)
group_stars_white = miru.graphics.PointSpriteGroup(
        stars_yellow.get_texture(), 16)
group_stars_cyan = miru.graphics.PointSpriteGroup(
        stars_cyan.get_texture(), 50, attenuate=True)

@window.event
def on_draw():
    window.clear()
    context.render()


def main():
    pyglet.app.run()


