# -*- test-case-name: miru.test.test_utils -*-
# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details

from twisted.python import filepath

from pyglet import gl
from pyglet import clock
import sys, os

def glvec(*args):
    return (gl.GLfloat * len(args))(*args)

## Debugging stuff

def show_caller(level):
    l = level + 1
    frame = sys._getframe(l)
    fname = frame.f_code.co_filename.split(os.path.sep)[-1]
    caller = frame.f_code.co_name
    lineno = frame.f_lineno
    print '%s %s %d' % (fname, caller, lineno)
showCaller = show_caller

_font_styles = {
    'mono' : ('VeraMono.ttf', 'Bitstream Vera Sans Mono'),
}
_defaultLoaded = False
def default_font(size=10, style='mono'):
    global _defaultLoaded
    from pyglet import font
    filename,face = _font_styles.get(style, _font_styles['mono'])
    if not _defaultLoaded:
        fontpath = filepath.FilePath(__file__).parent().child('data').child(filename)
        font.add_file(fontpath.path)
        _defaultLoaded = True
    return font.load(face, size)
defaultFont = default_font

def add_fps_display(context=None):
    if not context:
        from miru.context import context
    f = defaultFont(14)
    clock_display = clock.ClockDisplay(font=f, color=(0.,1.,0.,0.5))
    context.osd.add_object(clock_display)
addFpsDisplay = add_fps_display

def flatten(l, _cycle_checks=None):
    """Flatten a list of lists.
    """
    _cycle_checks = _cycle_checks or []
    newl = []
    for item in l:
        if isinstance(item, list):
            if item in _cycle_checks:
                raise ValueError, 'Cannot flatten cyclic lists'
            _cycle_checks.append(item)
            newl.extend(flatten(item, _cycle_checks))
        else:
            newl.append(item)
    return newl


def color(c0, c1=None, c2=None, c3=None):
    """Convert a color of any of the following formats RGB, RGBA, L, LA 
    from integral form in range [0,255] to floating point represetation
    in range [0.0, 1.0].  Internally, miru always uses the latter, so
    if you prefer the former, this function is your friend.
    """
    if c1 is None and c2 is None and c3 is None:
        return c0 / 255.
    if c2 is None and c3 is None:
        return (c0 / 255., c1 / 255.)
    if c3 is None:
        return (c0 / 255., c1 / 255., c2 / 255.)
    return (c0 / 255., c1 / 255., c2 / 255., c3 / 255.)

def point_in_rect(x, y, rx, ry, width, height):
    """2D point inside rectangle check
    """
    return (x >= rx) and (x <= rx + width) and\
           (y >= ry) and (y <= ry + height)

def change_cursor(style, window):
    cursor = window.get_system_mouse_cursor(style)
    window.set_mouse_cursor(cursor)


def scale_texcoords(texcoords, scale=1, format=2):
    c = list
    if type(texcoords) is tuple:
        c = tuple
    coords = []
    for i in range(0, len(texcoords), format):
        if format == 2:
            coords.extend(texcoords[i:i+format] + c((0,scale)))
        else:
            coords.extend(texcoords[i:i+format] + c((scale,)))
    return tuple(coords)

